#  CkTable
# :
#   
#     -textvar
#   
#     -width
#     -height
#     -titlerows
#     -yscroll
#     -state
#
# :
#   %W width [ ] ...
#   %W tag col|row [| {    -configure}]
#   %W cget option
#   %W configure {options val} ...
#   %W reload -         CkTable   row & col  
#

package provide CkTable 0.1

package require Ck

namespace eval CkTable {

proc CkTable { frame args } {
  variable CkTable
  
  array unset CkTable $frame,*
  frame $frame -takefocus 0
  bind $frame <FocusIn> {
    set w %W.row0.col0
    if [winfo exists $w] {
      focus $w
    }
  }  
  
  #   
  set CkTable(options) { -textvar -width -height -titlerows -yscroll -state }
  
  #  
  foreach { var val } $args {
    if { ![regexp " $var " $CkTable(options)] } {
      error "Bad option \"$var\": must be $CkTable(options)"
    }
    set CkTable($frame,opt,$var) $val
  }
  if ![info exists CkTable($frame,opt,-width)]      { set CkTable($frame,opt,-width) 5 }
  if ![info exists CkTable($frame,opt,-height)]     { set CkTable($frame,opt,-height) 10 }
  if ![info exists CkTable($frame,opt,-titlerows)]  { set CkTable($frame,opt,-titlerows) 0 }
  if ![info exists CkTable($frame,opt,-state)]      { set CkTable($frame,opt,-state) normal }
  if ![info exists CkTable($frame,opt,-textvar)]    { error "Need options -textvar" }
  
  global $CkTable($frame,opt,-textvar)


  set CkTable($frame,setarray,pozition) 0
  set CkTable($frame,tags)              ""
  set CkTable($frame,tags,width)        ""

  Bind:CkTable
  Bind:EntryCkTable

  #  
  PackTable $frame

  proc $frame { args } {
    global CkTable::CkTable
    
    set frame [lindex [info level 0] 0]
    set cmd   [lindex $args 0]
    switch -- $cmd {
      width {
        set arg [lrange $args 1 end]
        foreach { col val } $arg {
          for { set row -$CkTable::CkTable($frame,opt,-titlerows) } { $row < $CkTable::CkTable($frame,opt,-height) } { incr row } {
            if [winfo exists $frame.row$row.col$col] {
              $frame.row$row.col$col configure -width $val
            }
          }
        }
        set CkTable($frame,tags,width) $arg
      }
      tag {
        set tip [lindex $args 1]
        set arg [lrange $args 2 end]
        switch -- $tip {
          col { 
            #foreach { col opts } $arg {
            #  for { set row 0 } { $row < $CkTable::CkTable($frame,opt,-height) } { incr row } {
            #    eval $frame.row$row.col$col configure $opts
            #  }
            #}
          }
          row {
            foreach { row opts } $arg {
              lappend CkTable($frame,tag,row,$row) $opts
            }
          }
          default { error "Bad option \"$tip\": must be col, row"}
        }

        #  
        set tag [lrange $args 1 end]
        if { [lsearch -exact $CkTable($frame,tags) $tag] == -1 } {
          lappend CkTable($frame,tags) $tag
        }
      }
      cget {
        set opt [lindex $args 1]
        if { ![regexp " $opt " $CkTable(options)] } {
          error "Bad option \"$opt\": must be $CkTable(options)"
        }            
        return $CkTable($frame,opt,$opt)
      }
      configure {
        set focus [focus]
        set arg [lrange $args 1 end]
        foreach { opt val } $arg {
          set flg 0
          if { $opt == "-width" } {
            set CkTable($frame,opt,-width) $val
            PackTable $frame
            ResetTags $frame
            set flg 1
          }
          if { $opt == "-height" } {
            set CkTable($frame,opt,-height) $val
            PackTable $frame
            ResetTags $frame
            set flg 1
          }
          if { $opt == "-textvar" } {
            
            set flg 1
          }
        }
        if { $flg == 0 } {
          error "Bad option \"$arg\": must be -width, -height"
        }
        if [winfo exists $focus] { 
          focus $focus
        } else {
          focus $frame.row0.col0
        }
      }
      reload {
        SetArray $frame
        SetTags $frame
      }
      default { error "Bad option \"$cmd\": must be width, tag, cget, configure, reload" }
    }
  }
  
  rename ::$frame ::CkTable($frame,frame)
  Scope $frame
  return $frame
}

proc ResetTags { frame } {
  variable CkTable  
  
  if { [llength $CkTable($frame,tags)] > 0 } {
    foreach { i } $CkTable($frame,tags) {
      set tip [lindex $i 0]
      set arg [lrange $i 1 end]
      eval $frame tag $tip $arg
    }
  }
  if { $CkTable($frame,tags,width) != "" } {
    eval $frame width $CkTable($frame,tags,width)
  }
}

proc PackTable { frame } {
  variable CkTable

  foreach cw [winfo children $frame] { destroy $cw }

  #   foreground  label
  pack [label $frame.l]
  set labelfg [option get $frame.l foreground Label]
  if { $labelfg == "" } { set labelfg cyan }
  destroy $frame.l
  
  #      title
  for { set row -$CkTable($frame,opt,-titlerows) } { $row < 0 } { incr row } {
    pack [frame $frame.row$row -takefocus 0]
    for { set col 0 } { $col < $CkTable($frame,opt,-width) } { incr col } {
      pack [entry $frame.row$row.col$col -state disabled -takefocus 0 -fg $labelfg -justify center \
              -textvar $CkTable($frame,opt,-textvar)($row,$col)] -side left
    }
  }
  #      
  for { set row 0 } { $row < $CkTable($frame,opt,-height) } { incr row } {
    pack [frame $frame.row$row]
    for { set col 0 } { $col < $CkTable($frame,opt,-width) } { incr col } {
      pack [entry $frame.row$row.col$col -textvar CkTable::CkTable($frame,$row,$col) -state $CkTable::CkTable($frame,opt,-state)] -side left
      #pack [label $frame.row$row.coll$col -text "|"] -side left
      bindtags $frame.row$row.col$col "$frame.row$row.col$col [winfo toplevel $frame] EntryCkTable CkTable"
      Bind:Entry $frame $frame.row$row.col$col $row $col
    }
  }

  update
  #   
  SetArray $frame
}

proc SetArray { frame } {
  variable CkTable
  global $CkTable($frame,opt,-textvar)

  ClearTableData $frame
  if { $CkTable($frame,setarray,pozition) < 0 } { set CkTable($frame,setarray,pozition) 0 }
  
  #    -yscroll
  if [info exists CkTable($frame,opt,-yscroll)] {
    set arrayrow [expr [llength [array names $CkTable($frame,opt,-textvar) "*\[0-9],\[0-9]*"]] / $CkTable($frame,opt,-width) - 1]
    if { $arrayrow > 0 } {
      set firstset  [expr $CkTable($frame,setarray,pozition).0 / $arrayrow]
      set secondset [expr ($CkTable($frame,setarray,pozition).0 + $CkTable($frame,opt,-height)) / $arrayrow]
      if { $firstset >= 0 && $secondset > 0 } {
        eval $CkTable($frame,opt,-yscroll) $firstset $secondset
      }
    }
    
  }
  
  #    -textvar   cktable-
  for { set row 0 } { $row < $CkTable($frame,opt,-height) } { incr row } {
    set textvarrow [expr $row + $CkTable($frame,setarray,pozition)]
    for { set col 0 } { $col < $CkTable($frame,opt,-width) } { incr col } {
      if [info exists $CkTable($frame,opt,-textvar)($textvarrow,$col)] {
        set CkTable($frame,$row,$col) [subst $$CkTable($frame,opt,-textvar)($textvarrow,$col)]
      }
    }
  }
}

proc SetTags { frame } {
  variable CkTable  

  #   
  for { set row 0 } { $row < $CkTable($frame,opt,-height) } { incr row } {
    for { set col 0 } { $col < $CkTable($frame,opt,-width) } { incr col } {
      if { [$frame.row$row.col$col cget -attr] == "reverse" } {
        $frame.row$row.col$col configure -bg black -fg white
      } else {
        $frame.row$row.col$col configure -bg black -fg white -attr normal
      }
    }
  }
  
  #  col 
  for { set t 0 } { $t < [llength $CkTable($frame,tags)] } { incr t } {
    set atags [lindex $CkTable($frame,tags) $t]
    if { [lindex $atags 0] == "col" } {
      set cols [lrange $atags 1 end]
      foreach { col opt } $cols {
        for { set row 0 } { $row < $CkTable($frame,opt,-height) } { incr row } {
          eval $frame.row$row.col$col configure $opt
        }
      }
      
    }
  }

  set poz $CkTable($frame,setarray,pozition)
  for { set row $poz } { $row < [expr $poz + $CkTable($frame,opt,-height)] } { incr row } {
    if [info exists CkTable($frame,tag,row,$row)] {
      foreach opts $CkTable($frame,tag,row,$row) {
        for { set col 0 } { $col < $CkTable::CkTable($frame,opt,-width) } { incr col } {
          #  
          set vrow [expr $row - $poz]
          eval $frame.row$vrow.col$col configure $opts
        }
      }
    }
  }
}

proc ClearTableData { frame } {
  variable CkTable

  for { set row 0 } { $row < $CkTable($frame,opt,-height) } { incr row } {
    for { set col 0 } { $col < $CkTable($frame,opt,-width) } { incr col } {
      set CkTable($frame,$row,$col) ""
    }
  }
}

proc Bind:Entry { frame w row col } {
  variable CkTable

  if { $col < [expr $CkTable($frame,opt,-width) - 1] } {
    bind $w <Right> {
      if { [%W index insert] == [%W index end] } {
        focus [ck_focusNext %W]
      }
    }
  }
  if { $col > 0 } {
    bind $w <Left> {
      if { [%W index insert] == 0 } {
        focus [ck_focusPrev %W]
      }
    }
  }
  if { $row < [expr $CkTable($frame,opt,-height) - 1] } { 
    bind $w <Down> [list focus $frame.row[expr $row + 1].col$col]
  }
  if { $row == [expr $CkTable($frame,opt,-height) - 1] } { 
    bind $w <Down> {
      #         
      focus [ck_focusNext %W]
      focus %W
      global CkTable::CkTable
      set frame [winfo parent [winfo parent %W]]
      incr CkTable::CkTable($frame,setarray,pozition)
      CkTable::SetArray $frame
    }
  }
  if { $row > 0 } { 
    bind $w <Up> [list focus $frame.row[expr $row - 1].col$col]
  }
  if { $row == 0 } { 
    bind $w <Up> {
      #         
      focus [ck_focusPrev %W]
      focus %W
      global CkTable::CkTable
      set frame [winfo parent [winfo parent %W]]
      incr CkTable::CkTable($frame,setarray,pozition) -1
      CkTable::SetArray $frame
    }
  }
  bind $w <FocusOut> [list CkTable::Bind:Entry:FocusOut $frame $row $col]
  if { $row == [expr $CkTable($frame,opt,-height) - 1] && 
       $col == [expr $CkTable($frame,opt,-width) - 1]} {
    bind $w <Linefeed> {
      #         
      focus [lindex [winfo children [winfo parent %W]] 0]
      global CkTable::CkTable
      set frame [winfo parent [winfo parent %W]]
      incr CkTable::CkTable($frame,setarray,pozition)
      CkTable::SetArray $frame
    }
  }
  bind $w <Down> +[list CkTable::SetTags $frame]
  bind $w <Up>   +[list CkTable::SetTags $frame]
}

#       
proc Bind:Entry:FocusOut { frame row col } {
  global CkTable::CkTable 
  global $CkTable::CkTable($frame,opt,-textvar)

  if [info exists CkTable::CkTable($frame,$row,$col)] {
    set $CkTable::CkTable($frame,opt,-textvar)([expr $row + $CkTable::CkTable($frame,setarray,pozition)],$col) $CkTable::CkTable($frame,$row,$col)
  }
}

#

proc Bind:CkTable { } {
}

proc Bind:EntryCkTable { } {
  bind EntryCkTable <FocusIn> {
    set chailds [winfo children [winfo parent %W]]
    foreach i $chailds {
      $i configure -attr reverse
    }
    if { [%W cget -state] != "disabled" } {
      %W selection from 0
      %W selection to end
      %W icursor end
    }
  }
  bind EntryCkTable <FocusOut> { 
    set chailds [winfo children [winfo parent %W]]
    foreach i $chailds {
      $i configure -attr normal
    }
    %W selection clear 
  }
  bind EntryCkTable <Linefeed>  { 
    set next [ck_focusNext %W] 
    #      
    if { "$next" > "%W" } {
      focus $next
    }
  }
  bind EntryCkTable <Left>      { ckEntrySetCursor %W [expr [%W index insert] - 1] }
  bind EntryCkTable <Right>     { ckEntrySetCursor %W [expr [%W index insert] + 1] }
  bind EntryCkTable <Home>      { ckEntrySetCursor %W 0 }
  bind EntryCkTable <End>       { ckEntrySetCursor %W end }
  bind EntryCkTable <Delete>    { 
    if [%W selection present] {
      %W delete sel.first sel.last
    } else {
      %W delete insert
    }
  }
  bind EntryCkTable <ASCIIDelete> {
    if [%W selection present] {
      %W delete sel.first sel.last
    } else {
      %W delete insert
    }
  }
  bind EntryCkTable <BackSpace> { ckEntryBackspace %W }
  bind EntryCkTable <Select>    { %W selection from insert }
  bind EntryCkTable <KeyPress>  { ckEntryInsert %W %A }
  bind EntryCkTable <Control-a> { ckEntrySetCursor %W 0 }
  bind EntryCkTable <Control-b> { ckEntrySetCursor %W [expr [%W index insert] - 1] }
  bind EntryCkTable <Control-d> { %W delete insert }
  bind EntryCkTable <Control-e> { ckEntrySetCursor %W end }
  bind EntryCkTable <Control-f> { ckEntrySetCursor %W [expr [%W index insert] + 1] }
  bind EntryCkTable <Control-h> { ckEntryBackspace %W }
  bind EntryCkTable <Control-k> { %W delete insert end }
  bind EntryCkTable <Control-t> { ckEntryTranspose %W }
  #   
  bind EntryCkTable <Tab> {
    global CkTable::CkTable
    set frame [winfo parent [winfo parent %W]]
    #    () 
    set row [expr $CkTable::CkTable($frame,opt,-height) -1]
    set col [expr $CkTable::CkTable($frame,opt,-width) -1]
    set focus [ck_focusNext $frame.row$row.col$col]
    focus $focus
  }
  #   
  bind EntryCkTable <BackTab> {
    global CkTable::CkTable
    set frame [winfo parent [winfo parent %W]]
    set focus [ck_focusPrev $frame]
    focus $focus
  }
}

proc Scope { wdg } {
  namespace export $wdg
  namespace inscope :: [list namespace import CkTable::$wdg]  
}

Scope CkTable
}




